import aicsimageio
import numpy as np
import logging
import typing
import tifffile
from aicsimageio import AICSImage, aics_image
from multipledispatch import dispatch


log = logging.getLogger(__name__)

"""
ImageUtils
------------------
Class that contains some static methods to help handle images
Channel= 'C'
MosaicTile= 'M'
Samples= 'S'
SpatialX= 'X'
SpatialY= 'Y'
SpatialZ= 'Z'
Time= 'T'

Two main functions params are
image - AICSImage ("STCZYX")
image_ndarray - np.ndarray ("CZYX")
"""


class ImageUtils:

    @staticmethod
    def imread(image: typing.Union[np.ndarray, str]) -> AICSImage:
        with AICSImage(image) as img:
            img.size_z
            return img

    @staticmethod
    def imsave(image_ndarray: np.ndarray, path: str):
        tifffile.imsave(path, image_ndarray, image_ndarray.shape)

    @staticmethod
    @dispatch(AICSImage, int)
    def get_channel(image: AICSImage, channel_index: int) -> np.ndarray:
        return image.get_image_data("CZYX")[channel_index:channel_index+1, :, :, :]

    @staticmethod
    @dispatch(np.ndarray, int)
    def get_channel(image_ndarray: np.ndarray, channel_index: int) -> np.ndarray:
        return image_ndarray[channel_index:channel_index+1, :, :, :]

    @staticmethod
    def add_channel(image_ndarray: np.ndarray, channel) -> np.ndarray:
        new_image_ndarray = np.append(image_ndarray, channel, axis=0)
        return new_image_ndarray

    @staticmethod
    def image_to_ndarray(image):
        return image.get_image_data("CZYX")

    @staticmethod
    def get_channel_names(image):
        return image.get_channel_names()

    """
    Normalize all values between 0 to max_value
    """
    @staticmethod
    def normalize(image_ndarray,max_value=255,dtype=np.uint8) -> np.ndarray:
        temp_image = image_ndarray-np.min(image_ndarray)
        return (((temp_image)/(np.max(temp_image)))*max_value).astype(dtype)
    
    """
    to_shape changes the image shape according to the shape recieved
    """
    @staticmethod
    def to_shape(image_ndarray, shape, rescale_z=None, min_shape=None,modulo=None)  -> np.ndarray:
        new_shape = shape
        if (min_shape is not None):
            new_shape = np.maximum(shape,min_shape)
        if (modulo is not None):
            new_shape = new_shape + (modulo - np.mod(new_shape,modulo))
        c_, z_, y_, x_ = new_shape
        c, z, y, x = image_ndarray.shape
        y_pad = (y_-y)
        x_pad = (x_-x)
        z_pad = (z_-z)
        c_pad = (c_-c)
        if (rescale_z is not None):
            scaled_a = image_ndarray[:,::rescale_z,:,:]
            z_pad = (z_ - scaled_a.shape[1])
        else:
            scaled_a = image_ndarray
        if (c_pad >= 0):        
            scaled_a = np.pad(scaled_a,((c_pad//2, c_pad//2 + c_pad%2),(0,0),(0,0),(0,0)),mode = 'constant')
        else:
            scaled_a = scaled_a[abs(c_pad//2):scaled_a.shape[0]+(c_pad//2 + c_pad%2),:,:,:]
            
        if (z_pad >= 0):        
            scaled_a = np.pad(scaled_a,((0,0),(z_pad//2, z_pad//2 + z_pad%2),(0,0),(0,0)),mode = 'constant')
        else:
            scaled_a = scaled_a[:,abs(z_pad//2):scaled_a.shape[1]+(z_pad//2 + z_pad%2),:,:]
            
        if (y_pad >= 0):        
            scaled_a = np.pad(scaled_a,((0,0),(0,0),(y_pad//2, y_pad//2 + y_pad%2),(0,0)),mode = 'constant')
        else:
            scaled_a = scaled_a[:,:,abs(y_pad//2):scaled_a.shape[2]+(y_pad//2 + y_pad%2),:]
            
        if (x_pad >= 0):        
            scaled_a = np.pad(scaled_a,((0,0),(0,0),(0,0),(x_pad//2, x_pad//2 + x_pad%2)),mode = 'constant')
        else:
            scaled_a = scaled_a[:,:,:,abs(x_pad//2):scaled_a.shape[3]+(x_pad//2 + x_pad%2)]                        
        return scaled_a
    
    @staticmethod
    def slice_image(image_ndarray:np.ndarray, x_idx:tuple,y_idx:tuple,z_idx:tuple)->np.ndarray:
        n_dim = len(image_ndarray.shape)
        slices = [slice(None)] * n_dim
        slices[n_dim - 3] = slice(x_idx[0], x_idx[1])
        slices[n_dim - 2] = slice(y_idx[0], y_idx[1])
        slices[n_dim - 1] = slice(z_idx[0], z_idx[1])
        slices = tuple(slices)
        sliced_image = image_ndarray[slices]
        return sliced_image
    
    @staticmethod
    def slice_image(image_ndarray:np.ndarray, indexes:list)->np.ndarray:
        n_dim = len(image_ndarray.shape)
        slices = [slice(None)] * n_dim
        for i in range(len(indexes)):
            slices[i] = slice(indexes[i][0],indexes[i][1])
        slices = tuple(slices)
        sliced_image = image_ndarray[slices]
        return sliced_image
    
    """
    mask_image gets image and mask_template it masks the image according to the template and duplicate it accordingly
    """
    @staticmethod
    def mask_image(image_ndarray,mask_template_ndarray) -> np.ndarray:
        mask_ndarray = mask_template_ndarray
        for i in range(int(image_ndarray.shape[0])-1):
            mask_ndarray = ImageUtils.add_channel(mask_ndarray,mask_template_ndarray)
        return np.where(mask_ndarray==1.0,image_ndarray,np.zeros(image_ndarray.shape))




import logging
from cell_imaging_utils.datasets_metadata.table.datasetes_metadata_abstract_table import DatasetsMetaDataAbstractTable
import pandas as pd
import typing
from multipledispatch import dispatch
from numpy import number

log = logging.getLogger(__name__)

"""
DatasetsMetaDataSCV
------------------
SCV implementation of DatasetMetadataAbstract

"""


class DatasetMetadataSCV(DatasetsMetaDataAbstractTable):

    def __init__(self, destenation, source=None) -> None:
        super().__init__(destenation, source)
        if (self.source is not None):
            self.data = pd.read_csv(self.source)
        else:
             self.data = pd.DataFrame([],columns=[])
    def create(self):
        self.data.to_csv(self.destenation)



import numpy as np
import logging
import typing
import matplotlib.pyplot as plt




log = logging.getLogger(__name__)

"""
PlotsUtils
------------------
Class that contains some static methods to help handle and output plots
"""


class PlotsUtils:

    @staticmethod
    def plot_image_heatmap(image: np.ndarray,title: str, xlabel: str, ylabel: str, save: typing.Union[str, None]=None) -> None:
        plt.figure()
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.title(title)
        # plt.colorbar()
        plt.imshow(image, cmap='seismic', interpolation='nearest')
        # cb = np.linspace(0.0,1.0,20)
        # cax = plt.axes(cb)
        plt.colorbar()
        if (save != None):
            plt.savefig(save)
        plt.show()




import logging
import pandas as pd
import typing
from multipledispatch import dispatch
from cell_imaging_utils.datasets_metadata.dict.datasetes_metadata_abstract_dict import DatasetsMetaDataAbstractDict

log = logging.getLogger(__name__)

"""
DatasetsMetaDataPickle
------------------
Pickle implementation of DatasetMetadataAbstract

"""


class DatasetMetadataPickle(DatasetsMetaDataAbstractDict):

    def __init__(self, destenation, source=None) -> None:
        super().__init__(destenation, source)
        if (self.source is not None):
            self.data = pd.read_pickle(self.source)
        else:
            self.data = pd.DataFrame([])

    def create(self):
        self.data.to_pickle(self.destenation)



import os
import logging
import numpy as np

from cell_imaging_utils.plots.plot_utils import PlotsUtils
from cell_imaging_utils.image.image_utils import ImageUtils

log = logging.getLogger(__name__)
results_save_dir = "{}\\tests\\results".format(os.getcwd())
images_save_dir = "{}\\tests\\images".format(os.getcwd())

organelle_name = "Endoplasmic-reticulum"
# organelle_name = "Golgi"
# organelle_name = "Microtubules"
# organelle_name = "Lysosome"
# organelle_name =  "Tight-junctions"
# organelle_name = "Mitochondria"
image_file_name_1 = "0_prediction_c0..tif"
# image_file_name_2 = "0_signal.tif" #"0_prediction_c0..tif"
# image_file_name_2 = "0_target.tif"
image_file_name_2 = "0_seg.tif"
# result_fig = "signal_target_slice31.png"
result_fig = "seg_prediction_slice31.png"
# seg_image_file_name = "image_list_test.csv"

if not os.path.exists(results_save_dir):
    os.makedirs(results_save_dir)

def test_plot_utils() -> None:
    
    image1 = ImageUtils.imread("{}\\{}\\{}".format(images_save_dir,organelle_name,image_file_name_1))
    image2 = ImageUtils.imread("{}\\{}\\{}".format(images_save_dir,organelle_name,image_file_name_2))
    # print(ImageUtils.get_channel_names(image))
    image1_ndarray = ImageUtils.normalize(ImageUtils.image_to_ndarray(image1),max_value=255,dtype=np.double)
    image2_ndarray = ImageUtils.normalize(ImageUtils.image_to_ndarray(image2),max_value=255,dtype=np.double)
    diff_image = np.abs(image1_ndarray - image2_ndarray)
    diff_image = ImageUtils.normalize(diff_image,max_value=255,dtype=np.double)
    diff_image = np.clip(diff_image,0,100)
    PlotsUtils.plot_image_heatmap(diff_image[0,31,:,:], organelle_name,"x","y",save="{}\\{}_{}".format(results_save_dir,organelle_name,result_fig))
    # # image1_ndarray = np.clip(image1_ndarray,50,150)
    # PlotsUtils.plot_image_heatmap(image1_ndarray[0,30,:,:], organelle_name,"x","y")
    # # image2_ndarray = np.clip(image2_ndarray,50,150)
    # PlotsUtils.plot_image_heatmap(image2_ndarray[0,30,:,:], organelle_name,"x","y")
    return None


test_plot_utils()


import logging
import typing
from abc import ABC, abstractmethod
import pandas as pd
from multipledispatch import dispatch
from cell_imaging_utils.datasets_metadata.datasetes_metadata_abstract import DatasetsMetaDataAbstract

from numpy import number

log = logging.getLogger(__name__)

"""
DatasetsMetaDataAbstractTable
------------------
Abstract method th handle datasets metadata
the data is handled as DataFrame of pandas
has table shape


"""

class DatasetsMetaDataAbstractTable(DatasetsMetaDataAbstract):
     
     def __init__(self,destenation,source=None) -> None:
          super().__init__(destenation,source)
     
     def get_shape(self)->tuple:
          return self.data.shape
     
     @dispatch(int,int,(int,str,object))
     def set_item(self,row,column,value)->None:
          self.data.iat[row,column]=value
     
     @dispatch(int,str,(int,str,object))
     def set_item(self,row,column,value)->None:
          self.data.at[row,column]=value
     
     @dispatch(int,int)
     def get_item(self,row:int,column:int)->typing.Any:
          return self.data.iloc[row].iat[column]

     @dispatch(int,str)
     def get_item(self,row:int,column:str)->typing.Union[str,number]:
          return self.data.iloc[row].at[column]
     
     def set_row(self,row:int,value:list)->None:
          self.data.iloc[row] = value
     
     def get_row(self,row)->list:
          return self.data.iloc[row]
     
     def add_row(self,value)->None:
          new_row = pd.DataFrame([value], columns=self.data.columns.to_list()) 
          self.data = self.data.append(new_row)
     
     @dispatch(int,(int,str))
     def set_column(self,column:int,value)->None:
          self.data.iloc[:,column] = value
          
     @dispatch(str,(int,str,list))
     def set_column(self,column:str,value)->None:
          self.data.loc[:,column] = value
     
     @dispatch(str)
     def get_column(self,column:str)->list:
          return self.data.loc[:,column]
     
     @dispatch(int)
     def get_column(self,column:int)->list:
          return self.data.iloc[:,column]
     
     def create_header(self,column_names)->None:
          self.data =self.data.reindex(column_names,axis="columns")
     
     def add_column(self,column_name:str,value:list=None)->None:
          self.data = self.data[column_name] = value
     
     
     
     
     
     
     
     


import logging
import typing
from abc import ABC, abstractmethod
import pandas as pd
from multipledispatch import dispatch
from cell_imaging_utils.datasets_metadata.datasetes_metadata_abstract import DatasetsMetaDataAbstract

from numpy import number

log = logging.getLogger(__name__)

"""
DatasetsMetaDataAbstractDict
------------------
Abstract method th handle datasets metadata
the data is handled as DataFrame of pandas
has table shape


"""

class DatasetsMetaDataAbstractDict(DatasetsMetaDataAbstract):
     
     def __init__(self,destenation,source=None) -> None:
          super().__init__(destenation,source)
     

     
     
     
     
     
     
     
     